#define _USE_MATH_DEFINES
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <iomanip>
#include <omp.h>
#include <sstream>

// Change the line below with the path in which the stdfin library
// is installed in your system (http://stdfin.org/download)
#include "stdfin/nintegrate/gauss_legendre.hpp"

using namespace std;

class Node {

        private:
        int name;
        long double theta, radius, radius_init, R_i;
        vector<Node*> adjNodeList;
        vector<long double> cmnNbrList;

        public:
        Node(int id) {name = id;}

        void setAngle(long double angle) {theta=angle;}
        long double getAngle() {return theta;}

        void setRadius(long double r) {radius=r;}
        long double getRadius() {return radius;}

        void setInitRadius(long double r) {radius_init = r;}
        long double getInitRadius() {return radius_init;}

        void set_R(long double R) {R_i = R;}
        long double get_R() {return R_i;}

        int getName() {return name;}

        void addAdjNode(Node *adj) {adjNodeList.push_back(adj);}

        void addCmnNbr(long double val) {cmnNbrList.push_back(val);}

        vector<Node*> getAdjNodeList() {return adjNodeList;}

        bool checkLink(Node *x) {
             for (int i = 0; i < adjNodeList.size(); i++) {
                  if (adjNodeList[i] == x) { return true; }
             }
             return false;
        }

        long double getCmnNbr(int i) { return cmnNbrList[i];}
};
class Graph {

	private:
	vector<Node*> nodeList;

	public:
	void addNewNode(Node *nNode) { nodeList.push_back(nNode); }

	Node* findNode(int name) {
	      for (int i = 0; i < nodeList.size(); i++) {
	           if (nodeList[i]->getName() == name) return nodeList[i];
	      }
	      return NULL;
	}

	vector<Node*> getNodeList() { return nodeList; }
};

long double Random() {
        long double r = ((long double)rand()/(long double)RAND_MAX);
        return r*2*M_PI;
}

long double m = 1.5, L = 1, gma = 2.1, T = 0.8, zeta = 1, N = 42426;

long double r_v_k_1;
long double r_v_k_2;
long double r_u_k_1;
long double r_u_k_2;
long double R_v_k;
long double R_u_k;
long double theta_v;
long double theta_u;

double f(long double x) {

     return (1/(2*M_PI))*(1/(1+exp((zeta/(2*T))*((1/zeta)*acosh(cosh(zeta*r_v_k_1)*cosh(zeta*r_v_k_2)-sinh(zeta*r_v_k_1)*sinh(zeta*r_v_k_2)*cos(M_PI-fabs(M_PI-fabs(x-theta_v))))-R_v_k))))*(1/(1+exp((zeta/(2*T))*((1/zeta)*acosh(cosh(zeta*r_u_k_1)*cosh(zeta*r_u_k_2)-sinh(zeta*r_u_k_1)*sinh(zeta*r_u_k_2)*cos(M_PI-fabs(M_PI-fabs(x-theta_u))))-R_u_k))));

}

int main(int argc, char *argv[]) {

        long double maxDeg = 0, t = 0, m_t, R_t, empirical_cn = 0;

        long double Beta=1/(gma-1);

        long double bar_k=2*(m+L);

        Graph *G = new Graph();

        Node *u, *v, *l;

	//Read the graph.
	ifstream str(argv[1]);

	if (str.is_open()) {

		string line; 
                int vertex1, vertex2;

		while (getline(str,line)) {

                        vertex1 = atoi(line.substr(0,int(line.find(" "))).c_str());
                        vertex2 = atoi(line.erase(0,int(line.find(" "))+1).c_str());

			u = G->findNode(vertex1);
			if (u == NULL) {
			    u = new Node(vertex1);
			    G->addNewNode(u);
			}

			v = G->findNode(vertex2);
			if (v == NULL) {
			    v = new Node(vertex2);
			    G->addNewNode(v);
			}

			u->addAdjNode(v);
			v->addAdjNode(u);

			if (u->getAdjNodeList().size() > maxDeg) maxDeg = u->getAdjNodeList().size();
			if (v->getAdjNodeList().size() > maxDeg) maxDeg = v->getAdjNodeList().size();
		}

		str.close();
	}

	cout.precision(15);

	srand((unsigned)time(NULL));

 	vector<Node*> nodes = G->getNodeList(); 

        Graph *model = new Graph();

        //Section of code that uses common neighbor statistics begins.-------------------------------------------------------

	Graph *model_CN = new Graph();

        for (int i = maxDeg; i >= 1; i--) {
             for (int j = 0; j < nodes.size(); j++) {
                  if (nodes[j]->getAdjNodeList().size() == i){
                        t++;
                        long double I_t=(1/(1-Beta))*(1-pow(t,-(1-Beta)));
                        long double L_t=(2*L*(1-Beta)/((2*Beta-1)*pow((1-pow(N,-(1-Beta))),2)))*(pow(N/t, (2*Beta-1))-1)*(1-pow(t, -(1-Beta)));
                        long double m_t=m+L_t;
                        u = nodes[j];
                        long double r_t = (2/zeta)*log(t);
                        u->setInitRadius(r_t);
                        long double R_t=r_t-(2/zeta)*log((2*T/sin(T*M_PI))*(I_t/m_t));
                        u->set_R(R_t);
                        if((m_t >= t-1) || (t == 1)) {
			    model_CN->addNewNode(u);
                        }
                  }
              }
        }

        vector<Node*> nodes_CN = model_CN->getNodeList();
        
        t = 0;

        for (int i = 0; i < nodes_CN.size(); i++) {

             v = nodes_CN[i];

             t++;

             model->addNewNode(v);

             if(t == 1) {
                //v->setAngle(Random());
                v->setAngle(2.60221362397685);
                cout<<v->getName()<<" "<<v->getAngle()<<"\n";
                continue;
             }

             vector<Node*> nodes2 = model->getNodeList();

             vector<Node*> nbrList = v->getAdjNodeList();

             //Find the empirical number of common neighbors between v and each old node u.

             for (int x = 0; x < nodes2.size()-1; x++) {
                  u = nodes2[x];
                  empirical_cn = 0;
                  for (int k = 0; k < nbrList.size(); k++) {
                       l = nbrList[k];
                       if(l != u) {
                         if(u -> checkLink(l)) {
                            empirical_cn++;
                         }
                      }
                  }
                  v->addCmnNbr(empirical_cn);
             }

             long double step = (long double)1/nodes_CN.size(), maxlog_L=-1000000000;

             theta_v = 0.00000001;

             while (theta_v <= 2*M_PI) {

                   long double log_L = 0;

                   for (int x = 0; x < nodes2.size()-1; x++) {

                        u = nodes2[x];
                        long double t_u=exp((zeta/2)*u->getInitRadius());
                        theta_u = u->getAngle();
                        empirical_cn=v->getCmnNbr(x);

                        //Find the expected number of common neighbors between u and v.
                        long double lambda = 0;

                        for (int k = 0; k < nodes.size(); k++) {

                             l = nodes[k];

                             if((l != u) && (l != v)) {

                                 long double r_l = l->getInitRadius();
                                 long double t_l=exp((zeta/2)*l->getInitRadius());

                                 if(t_l > t) {//l came after v.
                                    r_v_k_2 = r_l;
                                    R_v_k = l->get_R();
                                    r_v_k_1 = Beta*v->getInitRadius()+(1-Beta)*r_l;
                                 } else {//v came after l.
                                    r_v_k_2 = v->getInitRadius();
                                    R_v_k = v->get_R();
                                    r_v_k_1 = Beta*r_l+(1-Beta)*v->getInitRadius();
                                 }

                                 if(t_l > t_u ) {//l came after u.
                                    r_u_k_2 = r_l;
                                    R_u_k = l->get_R();
                                    r_u_k_1 = Beta*u->getInitRadius()+(1-Beta)*r_l;
                                 } else {//u came after l.
                                    r_u_k_2 = u->getInitRadius();                                                                                                                       R_u_k = u->get_R();
                                    r_u_k_1 = Beta*r_l+(1-Beta)*u->getInitRadius();
                                 }

                                 long double prob = stdfin::gauss_legendre_48(f, 0, 2*M_PI);
                                 lambda = lambda + prob;

                              }//Endif.

                        }//Endfor, for all possible neighbors k.

                        long double tmp = empirical_cn;
                        long double log_fact = 0.0;

                        while(tmp >= 1) {
                             log_fact = log_fact + log(tmp);
                             tmp--;
                        }

                        long double log_prob_cn=-lambda+empirical_cn*log(lambda)-log_fact;
                        log_L=log_L+log_prob_cn;

                   }//Endfor, for all old nodes.

                   /*if(t==2) {
                      cout<<theta_v<<" "<<exp(log_L)<<"\n";
                   }*/

                   if (log_L >= maxlog_L) {
                       v->setAngle(theta_v);
                       maxlog_L = log_L;
                   }

                   theta_v = theta_v+step;

              }//Endwhile, for all angles.
  
             //if(t==2) {
                 cout <<v->getName()<<" "<<v->getAngle()<<"\n";
             //    exit(-1);
             //}

       }//Endfor, for all initial nodes. 

       //Section of code that uses common neighbor statistics ends.-------------------------------------------------------
  
       //Now, embed all nodes as usual using link-based ML. Skip nodes for which angular coordinates are known (i.e., computed aboved.)

       //exit(-1);
       long double t_CN = t;

       t = 0;

       //If you want to embed all nodes with degree some k' >=k, just change the 1 below to the k you want.
       for (int i = maxDeg; i >= 2; i--) {
	     for (int j = 0; j < nodes.size(); j++) {
                  if (nodes[j]->getAdjNodeList().size() == i) {

			    t++;

                            if(t <= t_CN) {
                               continue;
                            }

			    u = nodes[j];

			    long double r_t = (2/zeta)*log(t);
			    u->setRadius(r_t);
		   	    u->setInitRadius(r_t);

			    model->addNewNode(u);

			    vector<Node*> nodes2 = model->getNodeList();
       
                	    for (int k = 0; k < nodes2.size(); k++) {
                                 v = nodes2[k];
                                 if(u != v) {
                                    v->setRadius(Beta*v->getInitRadius() + (1-Beta)*r_t);
                       	         }
                            }

                            long double I_t=(1/(1-Beta))*(1-pow(t,-(1-Beta)));
                            long double L_t=(2*L*(1-Beta)/((2*Beta-1)*pow((1-pow(N,-(1-Beta))),2)))*(pow(N/t, (2*Beta-1))-1)*(1-pow(t, -(1-Beta)));
                            m_t=m+L_t;

                            R_t=r_t-(2/zeta)*log((2*T/sin(T*M_PI))*(I_t/m_t));
                             
                            u->set_R(R_t);

			    long double step = (long double)1/(t), maxlog_L=-1000000000;
                            long double theta=0;

                   	     while (theta <= 2*M_PI) {

			         long double theta_t = theta, log_L=0;

                                 #pragma omp parallel for private(v) reduction(+:log_L)
				 for (int k = 0; k < nodes2.size(); k++) {

				      v = nodes2[k];   

				      if (v != u){

				 	 long double theta_s = v->getAngle();
             				 long double r_s = v->getRadius();

                                         long double dtheta = fabs(theta_t - theta_s);
				       	 if (dtheta > M_PI) {
	 				    dtheta = 2*M_PI - dtheta;
					 }

					 long double x_st;
                                         if(dtheta==0) {
                                            x_st=fabs(r_t-r_s);
                                         } else { 
                                            x_st=(1/zeta)*acosh((cosh(zeta*r_s)*cosh(zeta*r_t))-(sinh(zeta*r_s)*sinh(zeta*r_t)*cos(dtheta)));
                                         }

					 long double P_st = 1/(1+exp((zeta/(2*T))*(x_st-R_t)));

         				 if (u->checkLink(v)) {
                                            log_L=log_L+log(P_st);
				 	 } else {
                                            log_L=log_L+log(1-P_st);
				         }
		   		     }
                                 }
			
		      	         if (log_L >= maxlog_L) {  
			             u->setAngle(theta_t);
				     maxlog_L = log_L;
                                 }
 
                                 theta=theta+step;
			    }
   
                            cout << t << " "<<u->getName() << " " << u->getAngle() << " " << u->getRadius() << "\n";

                    }//Endif.

	     }//Endfor (for all nodes with degree i).

             //This section runs correction steps as described in the paper. Put this section in comments if you don't need correction steps.
             if ((i==60) || (i==40) || (i==20) || (i==10)){

                  cout<<"Running correction step for all nodes with degree greater than "<<i<<"\n";

                  //We repeat each correction step a number of times, equal here to the average node degree (fewer times can also be beneficial.)
                  for (int round = 1 ; round <= bar_k ; round++) {

                        cout<<"round "<<round<<"\n";

                        vector<Node*> nodes2 = model->getNodeList();

                        for (int k = 0; k < nodes2.size(); k++) {

                             u = nodes2[k];

                             long double step = (long double)1/(t), maxlog_L=-1000000000;
                             long double theta=0;

                              while (theta <= 2*M_PI) {

                                     long double theta_u = theta, log_L=0;

                                     #pragma omp parallel for private(v) reduction(+:log_L)
                                     for (int l = 0; l < nodes2.size(); l++) {

                                          v = nodes2[l];

                                          if (v != u) {

                                              long double theta_v = v->getAngle();
                                              long double dtheta = fabs(theta_v - theta_u);
                                              if (dtheta > M_PI) {
                                                  dtheta = 2*M_PI - dtheta;
                                              }

                                              long double x_vu;
                                              long double R_init;

                                              if(u->getInitRadius() > v->getInitRadius()) {//u came after v.  

                                                 long double r_u=u->getInitRadius();
                                                 long double r_v=Beta*v->getInitRadius() + (1-Beta)*r_u;
                                                 if(dtheta==0) {
                                                   x_vu=fabs(r_u-r_v);
                                                 } else {
                                                   x_vu=(1/zeta)*acosh((cosh(zeta*r_v)*cosh(zeta*r_u))-(sinh(zeta*r_v)*sinh(zeta*r_u)*cos(dtheta)));
                                                 } 
                                                 R_init=u->get_R();

                                               } else {

                                                  long double r_v=v->getInitRadius();
                                                  long double r_u=Beta*u->getInitRadius() + (1-Beta)*r_v;
                                                  if(dtheta==0) {
                                                    x_vu=fabs(r_u-r_v);
                                                  } else {
                                                    x_vu=(1/zeta)*acosh((cosh(zeta*r_v)*cosh(zeta*r_u))-(sinh(zeta*r_v)*sinh(zeta*r_u)*cos(dtheta)));
                                                  }
                                                  R_init=v->get_R();

                                               }

                                               long double P_uv = 1/(1+exp((zeta/(2*T))*(x_vu-R_init)));

                                               if (u->checkLink(v)) {
                                                    log_L=log_L+log(P_uv);
                                                } else {
                                                    log_L=log_L+log(1-P_uv);
                                                }
                                          }
                                     }

                                     if (log_L >= maxlog_L) {
                                         u->setAngle(theta_u);
                                         maxlog_L = log_L;
                                     } 

                                     theta=theta+step;
                              }

                        }//Endfor (pick next node).
       
                  }//Endfor round.

             }//Endif for correction step. 

   }//Endfor (pick next node degree). 

  //Output the node coordinates in the form: node_id angular_coordinate radial_coordinate.
  ofstream out;
  out.precision(15);
  out.open("coordinates_embedding.txt");
  vector<Node*> nodes2 = model->getNodeList();
  for (int j = 0; j < nodes2.size(); j++) {
       u = nodes2[j];
       out << u->getName() << " " << u->getAngle() << " " << u->getRadius() << "\n";
  }
  out.close();

}
